from musicscore.dtd.dtd import Sequence, GroupReference, Element
from musicscore.musicxml.attributes.document_attributes import DocumentAttributes
from musicscore.musicxml.attributes.measure_attributes import MeasureAttributes
from musicscore.musicxml.attributes.part_attributes import PartAttributes
from musicscore.musicxml.groups.musicdata import MusicData, Attributes
from musicscore.musicxml.elements.scoreheader import ScoreHeader
from musicscore.musicxml.elements.xml_element import XMLElement

"""
	<xs:element name="score-partwise" block="extension substitution" final="#all">
		<xs:complexType>
			<xs:sequence>
				<xs:group ref="score-header"/>
				<xs:element name="part" maxOccurs="unbounded">
					<xs:complexType>
						<xs:sequence>
							<xs:element name="measure" maxOccurs="unbounded">
								<xs:complexType>
									<xs:group ref="music-data"/>
									<xs:attributeGroup ref="measure-attributes"/>
								</xs:complexType>
							</xs:element>
						</xs:sequence>
						<xs:attributeGroup ref="part-attributes"/>
					</xs:complexType>
				</xs:element>
			</xs:sequence>
			<xs:attributeGroup ref="document-attributes"/>
		</xs:complexType>
	</xs:element>
"""


class Measure(XMLElement, MeasureAttributes):
    _DTD = Sequence(
        GroupReference(MusicData)
    )

    def __init__(self, number=None, *args, **kwargs):
        super().__init__(tag='measure', number=number, *args, **kwargs)


class Part(XMLElement, PartAttributes):
    _DTD = Sequence(
        Element(Measure, max_occurrence=None)
    )

    def __init__(self, id, *args, **kwargs):
        super().__init__(tag='part', id=id, *args, **kwargs)


class Score(XMLElement, DocumentAttributes):
    """The score-partwise element is the root element for a partwise MusicXML score. It includes a score-header group
    followed by a series of parts with measures inside. The document-attributes attribute group includes the version
    attribute.
    """
    _DTD = Sequence(
        GroupReference(ScoreHeader),
        Element(Part, max_occurrence=None)
    )

    def __init__(self, *args, **kwargs):
        super().__init__(tag='score-partwise', *args, **kwargs)

    def write(self, path):
        xmlversion = '<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n'
        doctype = '<!DOCTYPE score-partwise PUBLIC "-//Recordare//DTD MusicXML {} Partwise//EN" "http://www.musicxml.org/dtds/partwise.dtd">\n'.format(
            self.version)

        if path[-4:] != '.xml':
            path += '.xml'
        output_file = open(path, 'w')
        output_file.write(xmlversion)
        output_file.write(doctype)
        output_file.write(self.to_string())
        output_file.close()


from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class LetterSpacing(AttributeAbstract):
    """
    The letter-spacing attribute specifies text tracking. Values are either "normal" or a number representing the number
    of ems to add between each letter. The number may be negative in order to subtract space. The default is normal,
    which allows flexibility of letter-spacing for purposes of text justification.
    """

    def __init__(self, letter_spacing=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('letter-spacing', letter_spacing, "TypeNumberOrNormal")


from musicscore.musicxml.attributes.color import Color
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.placement import Placement
from musicscore.musicxml.attributes.position import Position
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeNumberLevel, TypeUpDown


class ComplexTypeArpeggiate(ComplexType, Position, Placement, Color, OptionalUniqueId):
    """The arpeggiate type indicates that this note is part of an arpeggiated chord. The number attribute can be used to
    distinguish between two simultaneous chords arpeggiated separately (different numbers) or together (same number).
    The up-down attribute is used if there is an arrow on the arpeggio sign. By default, arpeggios go from the lowest
    to highest note."""

    def __init__(self, tag, number=1, direction='up', *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)
        self.number = number
        self.direction = direction

    @property
    def number(self):
        return self.get_attribute('number')

    @number.setter
    def number(self, value):
        if value is None:
            self.remove_attribute('number')
        else:
            TypeNumberLevel(value)
            self._ATTRIBUTES.insert(0, 'number')
            self.set_attribute('number', value)

    @property
    def direction(self):
        return self.get_attribute('direction')

    @direction.setter
    def direction(self, value):
        if value is None:
            self.remove_attribute('direction')
        else:
            TypeUpDown(value)
            self._ATTRIBUTES.insert(0, 'direction')
            self.set_attribute('direction', value)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class OptionalUniqueId(AttributeAbstract):
    """
    The optional-unique-id attribute group allows an element to optionally specify an ID that is unique to the entire
    document. This attribute group is not used for a required id attribute, or for an id attribute that specifies an
    id reference.
    """

    def __init__(self, id=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('id', id, 'ID')



from unittest import TestCase

from musicscore.musicxml.types.complextypes.attributes import Divisions


class TestDivisions(TestCase):

    def setUp(self):
        self.divisions = Divisions(1)

    def test_divisions(self):
        result = '''<divisions>1</divisions>
'''
        self.assertEqual(self.divisions.to_string(), result)
        self.divisions.value = 2
        result = '''<divisions>2</divisions>
'''
        self.assertEqual(self.divisions.to_string(), result)

        self.divisions.text = 3
        result = '''<divisions>3</divisions>
'''
        self.assertEqual(self.divisions.to_string(), result)




from unittest import TestCase
import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        sf = SimpleFormat(quarter_durations=[4, 4, 3, 2, 3, 4, 5, 6, 1, 2])
        self.score.set_time_signatures(quarter_durations=sf.quarter_duration)

        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)

        result_path = path + '_test_1'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_2(self):
        sf = SimpleFormat(quarter_durations=[4, 4, 3, 2, 3, 4, 5, 6, 1, 2])
        self.score.set_time_signatures(times={1: (3, 4), 5: (2, 4)})

        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)

        result_path = path + '_test_2'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_3(self):
        sf = SimpleFormat(quarter_durations=[4, 4, 3, 2, 3, 4, 5, 6, 1, 2])
        self.score.set_time_signatures(quarter_durations=[8, 3, 5, 4, 11, 3])

        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)

        result_path = path + '_test_3'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_4(self):
        sf = SimpleFormat(quarter_durations=[4, 4, 3, 2, 3, 4, 5, 6, 1, 2])
        self.score.set_time_signatures(quarter_durations=[8, 3, 5, 4, 11, 3], times={1: (3, 4)})

        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)

        result_path = path + '_test_4'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)



from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.printstyle import PrintStyleAlign
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeSmuflCodaGlyphNameType


class ComplexTypeCoda(ComplexType, PrintStyleAlign, OptionalUniqueId):
    """The coda type is the visual indicator of a coda sign. The exact glyph can be specified with the smufl attribute.
    A sound element is also needed to guide playback applications reliably.
    		<xs:attribute name="smufl" type="smufl-coda-glyph-name"/>"""

    def __init__(self, tag, smufl=None, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)
        self.smufl = smufl

    @property
    def smufl(self):
        return self.get_attribute('smufl')

    @smufl.setter
    def smufl(self, value):
        if value is None:
            self.remove_attribute('smufl')
        else:
            TypeSmuflCodaGlyphNameType(value)
            self._ATTRIBUTES.insert(0, 'smufl')
            self.set_attribute('smufl', value)



from unittest import TestCase

from musicscore.musicxml.elements.note import Type


class Test(TestCase):
    def test_1(self):
        type = Type('quarter', size='full')
        result = """<type size="full">quarter</type>
"""
        self.assertEqual(type.to_string(), result)


import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treechordflags3 import NoiseFlag3
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        xml_path = path + '_test_1.xml'
        durations = [0.25, 0.75, 0.5, 0.5, 1, 1.25, 1.5, 2, 2.5, 3, 3.5, 4, 4.5]
        sf = SimpleFormat(quarter_durations=durations)
        for chord in sf.chords:
            chord.add_flag(NoiseFlag3())

        sf.to_stream_voice().add_to_score(self.score)
        self.score.write(xml_path)
        # TestScore().assert_template(xml_path)



from musicscore.dtd.dtd import Sequence, Element
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import String


class VirtualLibrary(XMLElement, String):
    """
    The virtual-library element indicates the virtual instrument library name.
    """
    _TAG = 'virtual-library'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class VirtualName(XMLElement, String):
    """
    The virtual-name element indicates the library-specific name for the virtual instrument.
    """
    _TAG = 'virtual-name'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class ComplexTypeVirtualInstrument(ComplexType):
    """
    The virtual-instrument element defines a specific virtual instrument used for an instrument sound.
    """
    _DTD = Sequence(
        Element(VirtualLibrary, min_occurrence=0),
        Element(VirtualName, min_occurrence=0)

    )

    def __init__(self, tag, value=None, *args, **kwargs):
        super().__init__(tag=tag, value=value, *args, **kwargs)



from unittest import TestCase
import os

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        sf = SimpleFormat(midis=[(60)], quarter_durations=[2.2])
        sf.chords[0].add_tie('start')
        v = sf.to_stream_voice(1)
        # print(v.chords[0].tie_types)
        v.add_to_score(self.score, 1, 1)
        ch = self.score.get_measure(1).get_part(1).chords[0]
        result_path = path + '_test_1'
        self.score.write(path=result_path)



import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treemeasure import TreeMeasure
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()
        sf = SimpleFormat(quarter_durations=[4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4])
        v = sf.to_stream_voice(1)
        v.add_to_score(self.score, 1, 1)

    def test_1(self):
        self.score.page_style.orientation = 'landscape'
        result_path = path + '_test_1'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_2(self):
        for index, measure in enumerate(self.score.get_children_by_type(TreeMeasure)):
            if index % 4 == 0:
                measure.add_system_break()
        self.score.page_style.orientation = 'landscape'
        self.score.page_style.system_distance = 150
        result_path = path + '_test_2'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_3(self):
        # self.score.page_style.format = 'landscape'
        self.score.get_measure(5).add_system_break()
        # p = self.score.get_measure(5).get_part(1).get_children_by_type(Print)[0]
        # s = p.add_child(SystemLayout())
        # s.add_child(SystemDistance(300))
        self.score.get_measure(10).add_system_break()
        self.score.get_measure(10).add_system_distance(200)
        self.score.get_measure(12).add_system_distance(200)
        # self.score.page_style.system_distance = 20

        result_path = path + '_test_3'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)



from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test(self):
        simpleformat = SimpleFormat(quarter_durations=4)
        voice = simpleformat.to_stream_voice(1)
        voice.add_to_score(self.score, 1, 1)
        # print(voice.chords)
        # print(self.score.to_string())


from math import log2

from musicscore.musicxml.elements.fullnote import Pitch, Rest
from musicscore.musicxml.elements.note import Notehead


class Midi(object):
    """"""
    # (stem, alter, octaveAdd)
    standard = {
        0: ('C', 0, 0),
        0.5: ('C', 0.5, 0),
        1: ('C', 1, 0),
        1.5: ('D', -0.5, 0),
        2: ('D', 0, 0),
        2.5: ('D', 0.5, 0),
        3: ('E', -1, 0),
        3.5: ('E', -0.5, 0),
        4: ('E', 0, 0),
        4.5: ('E', 0.5, 0),
        5: ('F', 0, 0),
        5.5: ('F', 0.5, 0),
        6: ('F', 1, 0),
        6.5: ('G', -0.5, 0),
        7: ('G', 0, 0),
        7.5: ('G', 0.5, 0),
        8: ('A', -1, 0),
        8.5: ('A', -0.5, 0),
        9: ('A', 0, 0),
        9.5: ('A', 0.5, 0),
        10: ('B', -1, 0),
        10.5: ('B', -0.5, 0),
        11: ('B', 0, 0),
        11.5: ('C', -0.5, 1)
    }

    flat = {
        0: ('C', 0, 0),
        0.5: ('D', -1.5, 0),
        1: ('D', -1, 0),
        1.5: ('D', -0.5, 0),
        2: ('D', 0, 0),
        2.5: ('E', -1.5, 0),
        3: ('E', -1, 0),
        3.5: ('E', -0.5, 0),
        4: ('E', 0, 0),
        4.5: ('F', -0.5, 0),
        5: ('F', 0, 0),
        5.5: ('G', -1.5, 0),
        6: ('G', -1, 0),
        6.5: ('G', -0.5, 0),
        7: ('G', 0, 0),
        7.5: ('A', -1.5, 0),
        8: ('A', -1, 0),
        8.5: ('A', -0.5, 0),
        9: ('A', 0, 0),
        9.5: ('B', -1.5, 0),
        10: ('B', -1, 0),
        10.5: ('B', -0.5, 0),
        11: ('B', 0, 0),
        11.5: ('C', -0.5, 1)
    }

    sharp = {
        0: ('C', 0, 0),
        0.5: ('C', 0.5, 0),
        1: ('C', 1, 0),
        1.5: ('C', 1.5, 0),
        2: ('D', 0, 0),
        2.5: ('D', 0.5, 0),
        3: ('D', 1, 0),
        3.5: ('D', 1.5, 0),
        4: ('E', 0, 0),
        4.5: ('E', 0.5, 0),
        5: ('F', 0, 0),
        5.5: ('F', 0.5, 0),
        6: ('F', 1, 0),
        6.5: ('F', 1.5, 0),
        7: ('G', 0, 0),
        7.5: ('G', 0.5, 0),
        8: ('G', 1, 0),
        8.5: ('G', 1.5, 0),
        9: ('A', 0, 0),
        9.5: ('A', 0.5, 0),
        10: ('A', 1, 0),
        10.5: ('A', 1.5, 0),
        11: ('B', 0, 0),
        11.5: ('B', 0.5, 0)
    }

    enharmonic_1 = {
        0: ('B', 1, -1),
        0.5: ('D', -1.5, 0),
        1: ('D', -1, 0),
        1.5: ('C', 1.5, 0),
        2: ('C', 2, 0),
        2.5: ('E', -1.5, 0),
        3: ('D', 1, 0),
        3.5: ('D', 1.5, 0),
        4: ('F', -1, 0),
        4.5: ('F', -0.5, 0),
        5: ('E', 1, 0),
        5.5: ('G', -1.5, 0),
        6: ('G', -1, 0),
        6.5: ('F', 1.5, 0),
        7: ('F', 2, 0),
        7.5: ('A', -1.5, 0),
        8: ('G', 1, 0),
        8.5: ('G', 1.5, 0),
        9: ('G', 2, 0),
        9.5: ('B', -1.5, 0),
        10: ('A', 1, 0),
        10.5: ('A', 1.5, 0),
        11: ('C', -1, 1),
        11.5: ('B', 0.5, 0)
    }

    enharmonic_2 = {
        0: ('D', -2, 0),
        1: ('B', 2, -1),
        2: ('E', 2, 0),
        3: ('F', -2, 0),
        4: ('D', 2, 0),
        5: ('G', -2, 0),
        6: ('E', 2, 0),
        7: ('A', -2, 0),
        9: ('B', -2, 0),
        10: ('C', -2, 1),
        11: ('A', 2, 0)
    }

    def __init__(self, value=None, accidental_mode='standard', note_head=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._value = None
        self._accidental_mode = None
        self._note_head = None
        self.value = value
        self.accidental_mode = accidental_mode
        self.notehead = note_head

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, v):
        if v is not None:
            if not isinstance(v, float) and not isinstance(v, int):
                raise TypeError('midi.value must be of type float or int not{}'.format(type(v)))
            if v < 16 and v != 0:
                raise ValueError('midi.value {} must be greater than 16'.format(v))
        self._value = v

    @property
    def accidental_mode(self):
        return self._accidental_mode

    @accidental_mode.setter
    def accidental_mode(self, value):
        permitted = ('standard', 'flat', 'sharp', 'enharmonic_1', 'enharmonic_2')
        if value not in permitted:
            raise TypeError('accidental_mode.value {} must be in {}'.format(value, permitted))
        self._accidental_mode = value

    @property
    def notehead(self):
        return self._note_head

    @notehead.setter
    def notehead(self, val):
        if val is not None and not isinstance(val, Notehead):
            val = Notehead(val)
        self._note_head = val

    def add_notehead_object(self, val):
        if not isinstance(val, Notehead):
            raise TypeError()
        self.notehead = val

    def add_notehead(self, val, **kwargs):
        self.add_notehead_object(Notehead(val, **kwargs))

    def get_pitch_name(self):
        if self.accidental_mode == 'standard':
            output = self.standard[self.value % 12]

        elif self.accidental_mode == 'enharmonic_1':
            output = self.enharmonic_1[self.value % 12]

        elif self.accidental_mode == 'enharmonic_2':
            output = self.enharmonic_2[self.value % 12]

        elif self.accidental_mode == 'flat':
            output = self.flat[self.value % 12]

        elif self.accidental_mode == 'sharp':
            output = self.sharp[self.value % 12]

        output = list(output)
        output[2] += (int(self.value // 12)) - 1
        if output[1] == 0:
            output[1] = None

        return output

    def get_pitch_rest(self):
        if self.value == 0:
            return Rest()
        else:
            return Pitch(*self.get_pitch_name())

    def get_flat(self):
        return Midi(value=self.value - 1, accidental_mode='flat')

    def get_sharp(self):
        return Midi(value=self.value + 1, accidental_mode='sharp')

    def transpose(self, val):
        self.value += val
        return self
        # return Midi(value=self.value + val, accidental_mode=self.accidental_mode)

    @property
    def octave(self):
        return int(self.value / 12) - 1

    def __lt__(self, other):  # For x < y
        return self.value < other.value

    def __le__(self, other):  # For x <= y
        return self.value <= other.value

    def __gt__(self, other):  # For x > y
        return self.value > other.value

    def __ge__(self, other):  # For x >= y
        return self.value >= other.value

    @property
    def __name__(self):
        pitch_step = self.get_pitch_name()[1]

        if not pitch_step:
            accidental = ''
        elif pitch_step == -1.5:
            accidental = 'b-'
        elif pitch_step == -1:
            accidental = 'b'
        elif pitch_step == -0.5:
            accidental = '-'
        elif pitch_step == 0.5:
            accidental = '+'
        elif pitch_step == 1:
            accidental = '#'
        elif pitch_step == 1.5:
            accidental = '#+'
        else:
            accidental = str(pitch_step)

        return "{}{}{}".format(self.get_pitch_name()[0], accidental, self.octave)

    def __deepcopy__(self, memodict={}, **kwargs):
        output = self.__class__(value=self.value, accidental_mode=self.accidental_mode, note_head=self.notehead,
                                **kwargs)
        return output


class MidiNote(Midi):
    _VALUE = 60

    def __init__(self, octave, accidental=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._accidental = None
        self._octave = None

        self.octave = octave
        self.accidental = accidental

    def _set_new_value(self):
        if self._get_accidental_value() == -1:
            self.accidental_mode = 'flat'
        if self._get_accidental_value() == 1:
            self.accidental_mode = 'sharp'
        self.value = (self._VALUE + self._get_accidental_value()) - (4 - self.octave) * 12

    def _get_accidental_value(self):
        if self.accidental is None:
            return 0
        if self.accidental in ('flat', 'f', 'b'):
            return -1
        if self.accidental in ('sharp', 's', '#'):
            return 1

    @property
    def octave(self):
        return self._octave

    @octave.setter
    def octave(self, val):
        self._octave = val
        self._set_new_value()

    @property
    def accidental(self):
        return self._accidental

    @accidental.setter
    def accidental(self, val):
        self._accidental = val
        self._set_new_value()

    def __deepcopy__(self, memodict={}, **kwargs):
        output = super().__deepcopy__(octave=self.octave, accidental=self.accidental, **kwargs)
        return output

    def __repr__(self):
        return "{} at {}".format(self.__name__, id(self))


class C(MidiNote):
    _VALUE = 60


class D(MidiNote):
    _VALUE = 62


class E(MidiNote):
    _VALUE = 64


class F(MidiNote):
    _VALUE = 65


class G(MidiNote):
    _VALUE = 67


class A(MidiNote):
    _VALUE = 69


class B(MidiNote):
    _VALUE = 71


def midi_to_frequency(midi, a4=440):
    try:
        midi = midi.value
    except AttributeError:
        pass

    f = 2 ** ((midi - 69) / 12) * a4
    return f


def frequency_to_midi(frequency, a4=440):
    m = 69 + 12 * log2(frequency / a4)
    return m



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class DashLength(AttributeAbstract):
    def __init__(self, dash_length=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('dash-length', dash_length, "TypeTenths")


class SpaceLength(AttributeAbstract):
    def __init__(self, space_length=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('space-length', space_length, "TypeTenths")


class DashedFormatting(DashLength, SpaceLength):
    """
    The dashed-formatting entity represents the length of dashes and spaces in a dashed line. Both the dash-length and
    space-length attributes are represented in tenths. These attributes are ignored if the corresponding line-type attribute is not dashed.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)




from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.printstyle import PrintStyleAlign
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeTypeSmulfSegnoGlyphName


class Smulf(TypeTypeSmulfSegnoGlyphName):
    def __init__(self, value, *args, **kwargs):
        super().__init__(value=value, *args, **kwargs)


class ComplexTypeSegno(ComplexType, PrintStyleAlign, OptionalUniqueId, Smulf):
    """The segno type is the visual indicator of a segno sign. The exact glyph can be specified with the smufl
    attribute. A sound element is also needed to guide playback applications reliably."""

    def __init__(self, tag, value=None, *args, **kwargs):
        super().__init__(tag=tag, value=value, *args, **kwargs)



from musicscore.dtd.dtd import Sequence, Element

from musicscore.musicxml.types.complextypes.pagelayout import ComplexTypePageLayout
from musicscore.musicxml.types.complextypes.stafflayout import ComplexTypeStaffLayout
from musicscore.musicxml.types.complextypes.systemlayout import ComplexTypeSystemLayout

from musicscore.dtd.dtd import GroupReference, Element, Sequence

from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.simple_type import TypeTenths


#
# class LeftMargin(XMLElement, TypeTenths):
#     _TAG = 'left-margin'
#
#     def __init__(self, value=None, *args, **kwargs):
#         super().__init__(tag=self._TAG, value=value, *args, **kwargs)
#
#
# class RightMargin(XMLElement, TypeTenths):
#     _TAG = 'right-margin'
#
#     def __init__(self, value=None, *args, **kwargs):
#         super().__init__(tag=self._TAG, value=value, *args, **kwargs)
#
#
# class TopMargin(XMLElement, TypeTenths):
#     _TAG = 'top-margin'
#
#     def __init__(self, value=None, *args, **kwargs):
#         super().__init__(tag=self._TAG, value=value, *args, **kwargs)
#
#
# class BottomMargin(XMLElement, TypeTenths):
#     _TAG = 'bottom-margin'
#
#     def __init__(self, value=None, *args, **kwargs):
#         super().__init__(tag=self._TAG, value=value, *args, **kwargs)
#
#
# """
# The left-right-margins group specifies horizontal margins in tenths.
# """
#
# LeftRightMargins = Sequence(
#     Element(LeftMargin),
#     Element(RightMargin)
# )
#
# """
# The all-margins group specifies both horizontal and vertical margins in tenths.
# """
# AllMargins = Sequence(
#     GroupReference(LeftRightMargins),
#     Element(TopMargin),
#     Element(BottomMargin)
#
# )


class PageLayout(ComplexTypePageLayout):
    _TAG = 'page-layout'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class SystemLayout(ComplexTypeSystemLayout):
    _TAG = 'system-layout'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


class StaffLayout(ComplexTypeStaffLayout):
    _TAG = 'staff-layout'

    def __init__(self, *args, **kwargs):
        super().__init__(tag=self._TAG, *args, **kwargs)


"""
The layout group specifies the sequence of page, system, and staff layout elements that is
                common to both the defaults and print elements.
"""
Layout = Sequence(
    Element(PageLayout, min_occurrence=0),
    Element(SystemLayout, min_occurrence=0),
    Element(StaffLayout, min_occurrence=0, max_occurrence=None)
)



from musicscore.dtd.dtd import Sequence, Element
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import IDREF, TypeMidi16, String, TypePercent, TypeRotationDegrees, \
    TypeMidi16384, TypeMidi128


class MidiChannel(XMLElement, TypeMidi16):
    """The midi-channel element specifies a MIDI 1.0 channel numbers ranging from 1 to 16."""
    _TAG = 'midi-channel'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class MidiName(XMLElement, String):
    """The midi-name element corresponds to a ProgramName meta-event within a Standard MIDI File."""
    _TAG = 'midi-name'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class MidiBank(XMLElement, TypeMidi16384):
    """The midi-bank element specified a MIDI 1.0 bank number ranging from 1 to 16,384."""
    _TAG = 'midi-bank'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class MidiProgram(XMLElement, TypeMidi128):
    """The midi-program element specifies a MIDI 1.0 program number ranging from 1 to 128."""
    _TAG = 'midi-program'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class MidiUnpitched(XMLElement, TypeMidi128):
    """For unpitched instruments, the midi-unpitched element specifies a MIDI 1.0 note number ranging from 1 to 128.
    It is usually used with MIDI banks for percussion. Note that MIDI 1.0 note numbers are generally specified from 0
    to 127 rather than the 1 to 128 numbering used in this element."""
    _TAG = 'midi-unpitched'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class Volume(XMLElement, TypePercent):
    """The volume element value is a percentage of the maximum ranging from 0 to 100, with decimal values allowed. This
    corresponds to a scaling value for the MIDI 1.0 channel volume controller."""
    _TAG = 'volume'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class Pan(XMLElement, TypeRotationDegrees):
    """The pan and elevation elements allow placing of sound in a 3-D space relative to the listener. Both are expressed
    in degrees ranging from -180 to 180. For pan, 0 is straight ahead, -90 is hard left, 90 is hard right, and -180 and
    180 are directly behind the listener."""
    _TAG = 'rotation-degrees'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class Elevation(XMLElement, TypeRotationDegrees):
    """:documentation>The elevation and pan elements allow placing of sound in a 3-D space relative to the listener.
     Both are expressed in degrees ranging from -180 to 180. For elevation, 0 is level with the listener, 90 is directly
     above, and -90 is directly below."""
    _TAG = 'elevation'

    def __init__(self, value=None, *args, **kwargs):
        super().__init__(tag=self._TAG, value=value, *args, **kwargs)


class ComplexTypeMidiInstrument(ComplexType):
    """The midi-instrument type defines MIDI 1.0 instrument playback. The midi-instrument element can be a part of
    either the score-instrument element at the start of a part, or the sound element within a part. The id attribute
    refers to the score-instrument affected by the change."""

    _ATTRIBUTES = []
    _DTD = Sequence(
        Element(MidiChannel, min_occurrence=0),
        Element(MidiName, min_occurrence=0),
        Element(MidiBank, min_occurrence=0),
        Element(MidiProgram, min_occurrence=0),
        Element(MidiUnpitched, min_occurrence=0),
        Element(Volume, min_occurrence=0),
        Element(Pan, min_occurrence=0),
        Element(Elevation, min_occurrence=0)
    )

    def __init__(self, tag, id_, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)
        self.id = id_

    @property
    def id(self):
        return self.get_attribute('id')

    @id.setter
    def id(self, value):
        if value is None:
            self.remove_attribute('id')
        else:
            IDREF(value)
            self._ATTRIBUTES.insert(0, 'id')
            self.set_attribute('id', value)



from unittest import TestCase

from musicscore.musicxml.types.complextypes.attributes import Time, Beats, BeatType, SenzaMisura


class TestTime(TestCase):

    def setUp(self):
        self.time = Time()
        self.time.add_child(Beats(3))
        self.time.add_child(BeatType(2))

    def test_time(self):
        result = '''<time>
  <beats>3</beats>
  <beat-type>2</beat-type>
</time>
'''
        self.assertEqual(self.time.to_string(), result)

    def test_print(self):
        self.time.print_object = 'no'
        result = '''<time print-object="no">
  <beats>3</beats>
  <beat-type>2</beat-type>
</time>
'''
        self.assertEqual(self.time.to_string(), result)

        self.time.print_object = None
        result = '''<time>
  <beats>3</beats>
  <beat-type>2</beat-type>
</time>
'''
        self.assertEqual(self.time.to_string(), result)

        self.time.print_object = 'yes'
        result = '''<time print-object="yes">
  <beats>3</beats>
  <beat-type>2</beat-type>
</time>
'''
        self.assertEqual(self.time.to_string(), result)
        with self.assertRaises(ValueError):
            self.time.print_object = 'bla'

    def test_senza_misura(self):
        with self.assertRaises(NotImplementedError):
            self.time.add_child(SenzaMisura())

    def test_multiple_time_signatures(self):
        time = Time()
        time.add_child(Beats(3))
        time.add_child(BeatType(4))
        time.add_child(Beats(1))
        time.add_child(BeatType(8))
        result = '''<time>
  <beats>3</beats>
  <beat-type>4</beat-type>
  <beats>1</beats>
  <beat-type>8</beat-type>
</time>
'''
        self.assertEqual(time.to_string(), result)



from musicscore.musicxml.elements.fullnote import Pitch
from musicscore.musicxml.elements.note import Note, Duration
from musicscore.musicxml.elements.partwise import Score, Part, Measure
from musicscore.musicxml.elements.scoreheader import PartList
from musicscore.musicxml.groups.musicdata import Attributes
from musicscore.musicxml.types.complextypes.attributes import Divisions, Time, Beats, BeatType, Clef
from musicscore.musicxml.types.complextypes.clef import Sign, Line
from musicscore.musicxml.types.complextypes.partlist import ScorePart
from musicscore.musicxml.types.complextypes.scorepart import PartName

import os

from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class TestPartwise(TestScore):
    def setUp(self):
        self.score = Score()
        self.score.version = '3.0'
        part_list = self.score.add_child(PartList())
        score_part = part_list.add_child(ScorePart(id='P1'))
        score_part.add_child(PartName(name="Oboe", print_object='yes'))

        part = self.score.add_child(Part(id='P1'))
        measure = part.add_child(Measure(number='1'))
        attributes = measure.add_child(Attributes())
        attributes.add_child(Divisions(1))
        time = attributes.add_child(Time())
        time.add_child(Beats(4))
        time.add_child(BeatType(4))
        clef = attributes.add_child(Clef())
        clef.add_child(Sign('G'))
        clef.add_child(Line(2))
        note = measure.add_child(Note())
        note.add_child(Pitch())
        note.add_child(Duration(4))

    def test_score_head(self):
        # print(self.score.to_string())
        self.score.write(path=path)
        self.assert_template(path)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract, TypeBeamValue, TypeBeamLevel
from musicscore.musicxml.attributes.color import Color
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from unittest import TestCase


class Repeater(AttributeAbstract):
    def __init__(self, repeater=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('repeater', repeater, 'TypeYesNo')


class Fan(AttributeAbstract):
    def __init__(self, fan=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('fan', fan, 'TypeFan')


class ComplexTypeBeam(ComplexType, TypeBeamValue, Repeater, Fan, Color, OptionalUniqueId):
    """
    documentation>Beam values include begin, continue, end, forward hook, and backward hook. Up to eight concurrent
    beams are available to cover up to 1024th notes. Each beam in a note is represented with a separate beam element,
    starting with the eighth note beam using a number attribute of 1.

    Note that the beam number does not distinguish sets of beams that overlap, as it does for slur and other elements.
    Beaming groups are distinguished by being in different voices and/or the presence or absence of grace and cue
    elements.

    Beams that have a begin value can also have a fan attribute to indicate accelerandos and ritardandos using fanned
    beams. The fan attribute may also be used with a continue value if the fanning direction changes on that note. The
    value is "none" if not specified.

    The repeater attribute has been deprecated in MusicXML 3.0. Formerly used for tremolos, it needs to be specified
    with a "yes" value for each beam using it.
    """

    def __init__(self, value, number=1, *args, **kwargs):
        super().__init__(tag='beam', value=value, *args, **kwargs)
        self.number = number

    @property
    def number(self):
        return self.get_attribute('number')

    @number.setter
    def number(self, value):
        if value is None:
            self.remove_attribute('number')
        else:
            TypeBeamLevel(value)
            self._ATTRIBUTES.insert(0, 'number')
            self.set_attribute('number', value)


class Test(TestCase):

    def setUp(self):
        self.beam = ComplexTypeBeam('begin', number=2, repeater='yes', fan='accel', color='#800080', id='beam1')

    def test_beam(self):
        result = '''<beam number="2" repeater="yes" fan="accel" color="#800080" id="beam1">begin</beam>
'''
        self.assertEqual(self.beam.to_string(), result)

    def run(self):
        self.setUp()
        self.test_beam()
        print('beam tested')
        print()



from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

import os

from musicscore.musicxml.types.complextypes.notations import Slur
from tests.score_templates.xml_test_score import TestScore

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        sf = SimpleFormat(quarter_durations=[2, 2])
        slur = sf.chords[0].add_slur('start')
        slur.line_type = 'dashed'
        sf.chords[1].add_slur('stop')
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_1.xml'
        self.score.write(xml_path)

        TestScore().assert_template(xml_path)

    def test_2(self):
        sf = SimpleFormat(quarter_durations=[2, 2], midis=[(60, 63), (72, 76)])
        slur = sf.chords[0].add_slur('start')
        slur.line_type = 'dashed'
        sf.chords[1].add_slur('stop')
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_2.xml'
        self.score.write(xml_path)

        TestScore().assert_template(xml_path)

    def test_3(self):
        sf = SimpleFormat(quarter_durations=[2, 2, 2, 2], midis=[60, 63, 66, 69])
        # slur_1_1 = Slur(type='start', number=1)
        # slur_1_2 = Slur(type='stop', number=1)
        sf.chords[0].add_slur(type='start', number=1)
        sf.chords[-1].add_slur(type='stop', number=1)

        # slur_2_1 = Slur(type='start', number=2, line_type='dashed')
        # slur_2_2 = Slur(type='stop', number=2)
        sf.chords[0].add_slur(type='start', number=2, line_type='dashed')
        sf.chords[2].add_slur(type='stop', number=2)

        # slur_3_1 = Slur(type='start', number=3, line_type='dotted')
        # slur_3_2 = Slur(type='stop', number=3)
        sf.chords[2].add_slur(type='start', number=3, line_type='dotted')
        sf.chords[-1].add_slur(type='stop', number=3)

        # slur_4_1 = Slur(type='start', number=4, line_type='wavy')
        # slur_4_2 = Slur(type='stop', number=4)
        sf.chords[1].add_slur(type='start', number=4, line_type='wavy')
        sf.chords[2].add_slur(type='stop', number=4)

        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_3.xml'
        self.score.write(xml_path)

        TestScore().assert_template(xml_path)



import os
from unittest import TestCase

from quicktions import Fraction

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treechord import TreeChord
from musicscore.musictree.treechordflags3 import TreeChordFlag3
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from tests.score_templates.xml_test_score import TestScore

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self):
        self.score = TreeScoreTimewise()
        self.score.add_measure()
        self.score.add_part()

    def test_1(self):
        midis = [61, 61, 62, 60, 63, 64, 65, 61]
        for midi in midis:
            self.score.add_chord(1, 1, TreeChord(midi, quarter_duration=0.5))

        self.score.get_measure(1).get_part(1)
        self.score.finish()

        result_path = path + '_test_1'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_2(self):
        midis = [60.0, 60.5, 61.0, 62.5, 64.0, 66.0, 68.0, 69.5, 71.0, 71.5, 72.0, 71.5, 71.0, 69.5, 68.0, 66.0, 64.0,
                 62.5, 61.0, 60.5]
        measure_number = 1
        for midi in midis:
            chord = TreeChord(midi, quarter_duration=0.5)
            chord.add_lyric(midi)
            self.score.add_chord(measure_number, 1, chord)
            remaining_duration = self.score.get_measure(measure_number).get_part(1).get_voice(1).remaining_duration
            if remaining_duration == 0:
                self.score.add_measure()
                measure_number += 1
        self.score.accidental_mode = 'modern'
        result_path = path + '_test_2'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_3(self):
        midis = [(61.0, 63), 61.0, 0, 62.0, 61, 61, 61, (62, 61)]
        measure_number = 1
        for midi in midis:
            chord = TreeChord(midi, quarter_duration=0.5)
            chord.add_lyric([m.value for m in chord.midis])
            self.score.add_chord(measure_number, 1, chord)
            remaining_duration = self.score.get_measure(measure_number).get_part(1).get_voice(1).remaining_duration
            if remaining_duration == 0:
                self.score.add_measure()
                measure_number += 1
        self.score.accidental_mode = 'modern'
        result_path = path + '_test_3'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)

    def test_4(self):
        simpleformat = SimpleFormat(midis=list(range(60, 68)))
        voice = simpleformat.to_stream_voice(2)
        voice.add_to_score(self.score, 1, 1)
        result_path = path + '_test_4'
        self.score.accidental_mode = 'modern'
        self.score.write(result_path)
        TestScore().assert_template(result_path=result_path)

    def test_5(self):
        simpleformat = SimpleFormat(midis=[71.5, 71.5, 72, 72, 71.5, 71.5], quarter_durations=6 * [0.5])
        voice = simpleformat.to_stream_voice(1)
        voice.add_to_score(self.score, 1, 1)
        result_path = path + '_test_5'
        self.score.accidental_mode = 'normal'
        self.score.write(result_path)
        TestScore().assert_template(result_path=result_path)

    def test_6(self):
        midis = [51.5, 51.5, 50.5, 48.5, 49.5, 48.5, 50.0, 50.0, 49.5, 49.0]
        durations = [Fraction(255, 56), Fraction(6525, 3136), Fraction(6075, 3136), Fraction(2475, 3136),
                     Fraction(2145, 3136), Fraction(2805, 3136), Fraction(1815, 3136), Fraction(65, 56),
                     Fraction(2015, 1568), Fraction(1625, 1568)]
        simpleformat = SimpleFormat(midis=midis, quarter_durations=durations)
        simpleformat.auto_clef()
        voice = simpleformat.to_stream_voice(1)
        voice.add_to_score(self.score, 1, 1)
        result_path = path + '_test_6'
        self.score.max_division = 7
        self.score.accidental_mode = 'modern'
        self.score.write(result_path)
        TestScore().assert_template(result_path=result_path)

    def test_7(self):
        # todo update_accidental does not work ...
        class TestFlag3(TreeChordFlag3):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def implement(self, chord):
                split = chord.split(2, 2)
                split[0].to_rest()
                for ch in split:
                    ch.update_type()
                    ch.update_dot()
                return split

        xml_path = path + '_test_7.xml'
        sf = SimpleFormat(midis=[61], quarter_durations=[4])
        sf.to_stream_voice().add_to_score(self.score, part_number=1)
        chord = sf.chords[0]
        chord.add_flag(TestFlag3())
        sf.to_stream_voice().add_to_score(self.score, part_number=2)
        self.score.write(xml_path)

    def test_8(self):
        midis = [60 + factor * 0.5 for factor in range(0, 25)]
        simple_format = SimpleFormat(midis=midis + midis[-1::-1][1:])
        for index, chord in enumerate(simple_format.chords):
            if index <= len(midis) - 1:
                chord.midis[0].accidental_mode = 'sharp'
            else:
                chord.midis[0].accidental_mode = 'flat'
        simple_format.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_8.xml'
        self.score.write(xml_path)



import os
from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise

path = str(os.path.abspath(__file__).split('.')[0])


class Test(TestCase):
    def setUp(self) -> None:
        self.simple_format = SimpleFormat(quarter_durations=[4])
        self.score = TreeScoreTimewise()

    def test_1(self):
        sf = self.simple_format
        sf.chords[0].add_dynamics('ff')
        sf.chords[0].remove_dynamics()
        self.assertIsNone(sf.chords[0].get_dynamics())

    def test_2(self):
        sf = self.simple_format
        sf.chords[0].add_words('word')
        sf.chords[0].add_dynamics('ff')
        sf.chords[0].remove_dynamics()
        sf.chords[0].add_dynamics('p')
        sf.to_stream_voice().add_to_score(self.score, part_number=1)
        xml_path = path + '_test_2.xml'
        self.score.write(xml_path)



from musicscore.musicxml.attributes.justify import Justify
from musicscore.musicxml.attributes.printstyle import PrintStyle


class GroupNameText(PrintStyle, Justify):
    """
    The group-name-text attribute group is used by the group-name and group-abbreviation elements. The print-style and
    justify attribute groups are deprecated in MusicXML 2.0 in favor of the new group-name-display and
    group-abbreviation-display elements.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



'''
	<xs:complexType name="backup">
		<xs:annotation>
			<xs:documentation></xs:documentation>
		</xs:annotation>
		<xs:sequence>
			<xs:group ref="duration"/>
			<xs:group ref="editorial"/>
		</xs:sequence>
	</xs:complexType>
'''
from musicscore.dtd.dtd import Sequence, GroupReference, Element
from musicscore.musicxml.groups.common import Editorial
from musicscore.musicxml.elements.note import Duration
from musicscore.musicxml.types.complextypes.complextype import ComplexType


class ComplexTypeBackup(ComplexType):
    """
    The backup and forward elements are required to coordinate multiple voices in one part, including music on multiple
    staves. The backup type is generally used to move between voices and staves. Thus the backup element does not
    include voice or staff elements. Duration values should always be positive, and should not cross measure boundaries
    or mid-measure changes in the divisions value.
    """
    _DTD = Sequence(
        Element(Duration),
        GroupReference(Editorial)
    )

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from musicscore.musicxml.attributes.textformatting import TextFormatting
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeAccidentalValue

"""
<xs:attribute name="smufl" type="smufl-accidental-glyph-name"/>
"""


class ComplexTypeAccidentalText(ComplexType, TypeAccidentalValue, TextFormatting):
    """ The accidental-text type represents an element with an accidental value and text-formatting attributes."""

    def __init__(self, tag, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract
from musicscore.musicxml.attributes.color import Color
from musicscore.musicxml.attributes.dahsedformatting import DashedFormatting
from musicscore.musicxml.attributes.linetype import LineType
from musicscore.musicxml.attributes.optional_unique_id import OptionalUniqueId
from musicscore.musicxml.attributes.position import Position
from musicscore.musicxml.types.complextypes.complextype import ComplexType
from musicscore.musicxml.types.simple_type import TypeNumberLevel, TypeWedgeType


class Spread(AttributeAbstract):
    def __init__(self, spread=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('spread', spread, 'TypeTenths')


class Niente(AttributeAbstract):
    def __init__(self, niente=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('niente', niente, 'TypeYesNo')


class ComplexTypeWedge(ComplexType, Spread, Niente, LineType, DashedFormatting, Position, Color, OptionalUniqueId):
    """The wedge type represents crescendo and diminuendo wedge symbols. The type attribute is crescendo for the start
    of a wedge that is closed at the left side, and diminuendo for the start of a wedge that is closed on the right
    side. Spread values are measured in tenths; those at the start of a crescendo wedge or end of a diminuendo wedge are
    ignored. The niente attribute is yes if a circle appears at the point of the wedge, indicating a crescendo from
    nothing or diminuendo to nothing. It is no by default, and used only when the type is crescendo, or the type is
    stop for a wedge that began with a diminuendo type. The line-type is solid by default."""

    def __init__(self, tag, type, *args, **kwargs):
        super().__init__(tag=tag, *args, **kwargs)
        self.type = type

    @property
    def type(self):
        return self.get_attribute('type')

    @type.setter
    def type(self, value):
        if value is None:
            self.remove_attribute('type')
        else:
            TypeWedgeType(value)
            self._ATTRIBUTES.insert(0, 'type')
            self.set_attribute('type', value)

    @property
    def number(self):
        return self.get_attribute('number')

    @number.setter
    def number(self, value):
        if value is None:
            self.remove_attribute('number')
        else:
            TypeNumberLevel(value)
            self._ATTRIBUTES.insert(0, 'number')
            self.set_attribute('number', value)



from unittest import TestCase

from musicscore.musicstream.streamvoice import SimpleFormat
from musicscore.musictree.treescoretimewise import TreeScoreTimewise
import os

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        sf = SimpleFormat(quarter_durations=[1, 1, 1, 1, 1])
        for chord in sf.chords:
            chord.add_articulation('accent')
            chord.add_articulation('tenuto')
        sf.chords[0].add_slur('start')
        sf.chords[-1].add_slur('stop')
        sf.to_stream_voice().add_to_score(self.score)
        xml_path = path + '_test_1.xml'
        self.score.write(xml_path)



from musicscore.musicxml.attributes.attribute_abstract import AttributeAbstract


class Id(AttributeAbstract):
    def __init__(self, id, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generate_attribute('id', id, "IDREF")



from unittest import TestCase

from musicscore.dtd.dtd import Sequence, Element, ChildOccurrenceDTDConflict, ChildTypeDTDConflict
from musicscore.musicxml.elements.xml_element import XMLElement
from musicscore.musicxml.exceptions import ChildAlreadyExists


class A(XMLElement):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag='a', *args, **kwargs)
        self.multiple = True


class B(XMLElement):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag='b', *args, **kwargs)


class C(XMLElement):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag='c', *args, **kwargs)


class D(XMLElement):
    """"""

    def __init__(self, *args, **kwargs):
        super().__init__(tag='d', *args, **kwargs)


class E(XMLElement):
    """"""
    _DTD=Sequence(
        Element(C),
        Element(A, max_occurrence=None),
        Element(B)
    )

    def __init__(self, *args, **kwargs):
        super().__init__(tag='e', *args, **kwargs)


class TestAddChildren(TestCase):

    def test_add_child(self):
        e = E()
        e.add_child(C())
        # with self.assertRaises(ChildTypeDTDConflict):
        #     e.add_child(D())
        # self.assertEqual(len(e.get_children()), 1)
        # self.assertEqual(type(e.get_children()[0]), C)

    def test_add_multiple_child(self):
        e = E()
        e.add_child(A())
        e.add_child(C())
        e.add_child(A())
        e.add_child(A())
        self.assertEqual(len(e.get_children_by_type(A)), 3)
        self.assertEqual(len(e.get_children_by_type(C)), 1)
        self.assertEqual(len(e.get_children_by_type(D)), 0)

    def test_non_multiple_child(self):
        e = E()
        e.add_child(C())
        with self.assertRaises(ChildOccurrenceDTDConflict):
            e.add_child(C())


    def test_sort_children(self):
        e = E()
        e.add_child(A())
        e.add_child(B())
        e.add_child(A())
        e.add_child(C())
        result = [C, A, A, B]
        self.assertEqual([type(child) for child in e.get_children()], result)

    # def test_access_child(self):
    #     e = E()
    #     e.add_child(A())
    #     e.add_child(B())
    #     e.add_child(A())
    #     e.add_child(C())
    #
    #     self.assertEqual([type(child) for child in e.a], [A, A])
    #     self.assertEqual(type(e.b), B)
    #     with self.assertRaises(AttributeError):
    #         print(e.d)



import os
from unittest import TestCase

from musicscore.musictree.treescoretimewise import TreeScoreTimewise
from musicscore.musicxml.groups.musicdata import Direction
from tests.score_templates.xml_test_score import TestScore
from musicscore.musicxml.types.complextypes.direction import DirectionType, Sound
from musicscore.musicxml.types.complextypes.directiontype import Metronome
from musicscore.musicxml.types.complextypes.metronome import BeatUnit, PerMinute

path = os.path.abspath(__file__).split('.')[0]


class Test(TestCase):
    def setUp(self) -> None:
        self.score = TreeScoreTimewise()

    def test_1(self):
        self.score.add_measure()
        self.score.add_part()
        part = self.score.get_measure(1).get_part(1)
        d = part.add_child(Direction())
        dt = d.add_child(DirectionType())
        m = dt.add_child(Metronome())
        m.add_child(BeatUnit('quarter'))
        m.add_child(PerMinute('100'))
        d.add_child(Sound(tempo=100))

        result_path = path + '_test_1'
        self.score.write(path=result_path)
        TestScore().assert_template(result_path=result_path)
